#include <pthread.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <assert.h>
#include <ctype.h>
char *strdup(const char *);
#if defined (PTHREAD_THREADS_MAX)
#   define MAXTHREADS PTHREAD_THREADS_MAX
#else
#   define MAXTHREADS 100
#endif
#define WAIT_TIME  2 /* 2 seconds default */
/******************* QUEUE Operations ***********************/
/********************** QUEUE Data Structures ********************/
typedef struct llist {
  struct llist *next;
  void *data;
} LLIST;

typedef struct lqueue {
  LLIST *head, *tail;
  pthread_mutex_t mut;
  pthread_cond_t cond;
  volatile int count;
  int wait_time; /* Number of seconds to wait on queue */
} LQUEUE;
/*
 * CreateQueue
 */
LQUEUE *CreateQueue(int wait_time)
{
  int rv;
  LQUEUE *new = malloc(sizeof(LQUEUE));
  if (new) {
    new->head = new->tail = NULL;
    rv = pthread_mutex_init(&new->mut, NULL);
    if (rv) {
      free(new);
      return NULL;
    }
    rv = pthread_cond_init(&new->cond, NULL);
    if (rv) {
      pthread_mutex_destroy(&new->mut);
      free(new);
      return NULL;
    }
    new->count = 0;
    new->wait_time = wait_time;
  }
  return new;
}
/*
 * QueueDestroy
 */
/* Assumes that queue is empty */
void QueueDestroy(LQUEUE *queue)
{
  pthread_mutex_destroy(&queue->mut);
  pthread_cond_destroy(&queue->cond);
  free(queue);
}
/*
 * AddDatatoQueue
 */
/* Always signal when adding data */
int AddDatatoQueue(void *data, LQUEUE *queue)
{
  LLIST *newlist = malloc(sizeof(LLIST));
  if (newlist == NULL)
    return -1;
  newlist->next = NULL;
  newlist->data = data;
  /* Acquire the queue */
  pthread_mutex_lock(&queue->mut);
  if (queue->count == 0) {
    queue->head = queue->tail = newlist;
  } else {
    queue->tail->next = newlist;
    queue->tail = newlist;
  }
  queue->count++;
  /* Release the queue and notify clients */
  pthread_mutex_unlock(&queue->mut);
  pthread_cond_signal(&queue->cond);
  return 0;
}
/*
 * RemoveDataFromQueue
 */
/* This function is used to simply remove data from queue
 * Called by a thread that has not already acquired the lock
 */
void *RemoveDataFromQueue(LQUEUE *queue)
{
  LLIST *this;
  void *data = NULL;
  pthread_mutex_lock(&queue->mut);
  if (queue->count > 0) {
    this = queue->head;
    if (--(queue->count) <= 0) {
      queue->head = queue->tail = NULL;
    } else {
      queue->head = this->next;
    }
    data = this->data;
    free(this);
  }
  pthread_mutex_unlock(&queue->mut);
  return data;
}
/*
 * RemoveDataFromQueueNolock
 */
/* This function is used to simply remove data from queue
 * Called by a waiting thread that must lock the queue first
 */
void *RemoveDataFromQueueNolock(LQUEUE *queue)
{
  LLIST *this;
  void *data = NULL;
  if (queue->count > 0) {
    this = queue->head;
    if (--(queue->count) <= 0) {
      queue->head = queue->tail = NULL;
    } else {
      queue->head = this->next;
    }
    data = this->data;
    free(this);
  }
  return data;
}
/************** END OF QUEUE OPERATIONS *******************/
/*********** General data structures *********************/

typedef struct {
  int tnum; /* Number assigned by creator */
  pthread_t tid; /* My thread id assigned by system */
  int verbose;
  int rv;       /* value I return */
  LQUEUE *queue; /* My queue */
  /* for servers, name of input file, for clients, pointer to
   * OFTYPE.
   */
  void *file; 
} TARG;

typedef struct {
  char *filename;
  pthread_mutex_t mut;
} OFTYPE;
OFTYPE ofile = {
  NULL, PTHREAD_MUTEX_INITIALIZER
};
/*********************** Global Data ***************************/
/* 
 * RunFlag:When zero, tells threads to quit.
 * Does not need to be protected by mutex
 */
static volatile RunFlag = 1; 
TARG threads[MAXTHREADS];

/*****************************************************************/
int IsWS(char *s)
{
  /* Bypass leading spaces */
  while(*s && isspace(*s)) ++s;
  return (*s == '\0');
}
/* This can side affect ibuf */
char *tdup(char *ibuf)
{
  char *sp;
  /* move to first non white */
  while(*ibuf && isspace(*ibuf))
    ++ibuf;
  if (*ibuf == '\0')
    return NULL;
  sp = ibuf + strlen(ibuf) -1; /* point to last char of string */
  /* decrement sp until it points to last non-white */
  while(isspace(*sp)) --sp;
  sp[1] = '\0'; 
  return strdup(ibuf);
}
/**********************************************
 * Server:
 * Read input file until end of file.
 * place each record on the queue, let a 
 */
static void *Server(void *argp)
{
  TARG *targ = argp;
  FILE *ifile;
  struct timespec abstime;
  char buf[256];
  char *data;
  int rv;
  int added = 0;
  if (targ->file == NULL) {
    fprintf(stderr, "S%03X:No file name\n", targ->tnum);
    targ->rv = -1;
    pthread_exit(targ);
  }
  if (targ->queue == NULL) {
    fprintf(stderr, "S%03X:No queue\n", targ->tnum);
    targ->rv = -1;
    pthread_exit(targ);
  }
  if (targ->verbose)
    fprintf(stderr, "S%03X:running\n", targ->tnum);
    
  ifile = fopen(targ->file, "r");
  while(fgets(buf, sizeof(buf), ifile)) {
    if (IsWS(buf))
      continue;
    data = tdup(buf); /* make a trimmed copy */
    if (data == NULL) {
      fprintf(stderr, "S%03X:Unable to allocate:%s\n", 
	      targ->tnum, 
	      strerror(errno));
      targ->rv = -1;
      pthread_exit(targ);
    }
    added++;
    AddDatatoQueue(data, targ->queue);
    rv = pthread_cond_signal(&targ->queue->cond);
    if (rv) {
      fprintf(stderr, "S%03X:Error on cond signal:%s\n", 
	    targ->tnum, strerror(rv));
    }
  }
  fclose(ifile);
  /* Now wait for queue to empty */
  rv = pthread_mutex_lock(&targ->queue->mut);
  if (rv) {
    fprintf(stderr, "S%03X:Error on queue mutex:%s\n", 
	    targ->tnum, strerror(rv));
    free(data);
    targ->rv = -1;
    return targ;
  }
  while(RunFlag && targ->queue->count > 0) {
    /* Recompute wait time */
    gettimeofday((struct timeval*)&abstime, NULL);
    abstime.tv_sec += targ->queue->wait_time; /* Time out after 10 seconds */
    abstime.tv_nsec = 0;
    rv = pthread_cond_timedwait(&targ->queue->cond,
				&targ->queue->mut,
				&abstime);
    if (rv && rv != ETIMEDOUT) {
      fprintf(stderr, "S%03X:Error on cond wait:%s\n", 
	      targ->tnum, strerror(rv));
      break;
    }
  }
  pthread_mutex_unlock(&targ->queue->mut);
  RunFlag = 0;
  pthread_cond_broadcast(&targ->queue->cond);
  if (targ->verbose)
    fprintf(stderr, "S%03X:Normal Exit, %d records queued\n", 
	    targ->tnum, added);
  targ->rv = 0;
  pthread_exit(targ);
}
/* writes data to file, frees data */
static int process_data(int tnum, OFTYPE *ofile, void *data)
{
  FILE *fp;
  int rv;
  fp = fopen(ofile->filename, "a+");
  if (fp == NULL) {
    fprintf(stderr, "C%03X:Error on open %s:%s\n", 
	    tnum, ofile->filename, strerror(errno));
    free(data);
    return -1;
  }
  rv = pthread_mutex_lock(&ofile->mut);
  if (rv) {
    fprintf(stderr, "C%03X:Error on file mutex:%s\n", 
	    tnum, strerror(rv));
    free(data);
    return -1;
  }
  fprintf(fp, "C%03X:%s\n", tnum, data);
  fclose(fp);
  pthread_mutex_unlock(&ofile->mut);
  free(data);
  return 0;
}
/* 
 * Wait on queue until either signalled or done.
 */
static void *Client(void *argp)
{
  int rv;
  pthread_t tid;
  TARG *targ;
  struct timespec abstime;
  LQUEUE *queue;
  void *data;
  int processed = 0;
  int timedout = 0;
  targ = (TARG *)argp;
  targ->rv = 0;
  queue = targ->queue;
  if (targ->verbose)
    fprintf(stderr, "C%03X:running acquiring\n", targ->tnum);
  rv = pthread_mutex_lock(&queue->mut);
  if (rv) {
    fprintf(stderr, "C%03X:Error on mutex:%s\n", 
	    targ->tnum, strerror(rv));
    targ->rv = -1;
    pthread_exit(targ);
  }
  if (targ->verbose)
    fprintf(stderr, "C%03X:running acquired\n", targ->tnum);
  /* We loop on the predicate. cond_timedwait can wakeup
   * for other reasons
   */
  while(RunFlag) {
    /* Remove elements from queue */
    while (targ->queue->count > 0) {
      pthread_cond_broadcast(&queue->cond);
      if ((data = RemoveDataFromQueueNolock(queue))) {
	pthread_mutex_unlock(&queue->mut);
	/* allow another thread to run */
	pthread_cond_signal(&queue->cond);
	++processed;
	process_data(targ->tnum, targ->file, data);
	/* Reacquire lock for the cond wait */
	pthread_mutex_lock(&queue->mut);
      }
    }
    /* Recompute wait time */
    gettimeofday((struct timeval*)&abstime, NULL);
    abstime.tv_sec += targ->queue->wait_time; /* Time out after n seconds */
    abstime.tv_nsec = 0;
    /* Recheck predicate before waiting */
    if (RunFlag == 0) {
      break;
    }
    rv = pthread_cond_timedwait(&queue->cond,
				&queue->mut,
				&abstime);
    if (rv && rv != ETIMEDOUT) {
      fprintf(stderr, "C%03X:Error on cond wait:%s\n", 
	      targ->tnum, strerror(rv));
      targ->rv = 0;
      break;
    } else if (rv == ETIMEDOUT) {
      rv = 0;
      timedout++;
    }
  }
  pthread_mutex_unlock(&queue->mut);
  if (targ->verbose) {
    if (rv == 0) {
      if (timedout)
	fprintf(stderr, "C%03X:Successful completion, %d records, %d timeouts\n", 
		targ->tnum, processed, timedout);
      else
	fprintf(stderr, "C%03X:Successful completion, %d records\n", 
		targ->tnum, processed);
    }
  }
  pthread_exit(targ);
}
static char *iam;
void usage(const char *iam)
{
  fprintf(stderr, "Usage: %s -s n -c n [ -v ] [ -w n ] input-file output-file where\n", iam);
  fprintf(stderr, "\t-s\tNumber of servers (must be > 0)\n");
  fprintf(stderr, "\t-c\tNumber of clients (must be > 0)\n");
  fprintf(stderr, "\t-w\tWait time for client queue\n");
  fprintf(stderr, "\t-v\tRun verbosely\n");
}

int main(argc, argv)
int argc; 
char **argv;
{
  int i;
  int fd, rv, c;
  int clients = 0;
  int servers = 0;
  char *ifile = NULL;
  int wait_time = WAIT_TIME;
  void *trv;
  struct timespec abstime;
  /* nthreads == number of child processes */
  int nthreads = 0;
  LQUEUE *queue;
  pthread_t tid;
  int thread_num = 0;
  int rq;
  int verbose = 0;

  if (iam = strrchr(argv[0], '/'))
    ++iam;
  else
    iam = argv[0];
  tid = pthread_self(); /* Get my threadid */

  while((c = getopt(argc, argv, "vc:s:w:")) != -1) {
    switch(c) {
    case 'v':
      verbose = 1;
      break;
    case 'c':
      clients = atoi(optarg);
      break;
    case 's':
      servers = atoi(optarg);
      break;
    case 'w':
      wait_time = atoi(optarg);
      break;
    default:
      usage(iam);
      return -1;
    }
  }
  if (optind < argc)
    ifile = argv[optind++];

  if (optind < argc)
    ofile.filename = argv[optind++];
  if (ifile == NULL || ofile.filename == NULL) {
    usage(iam);
  }
  queue = CreateQueue(wait_time);
  assert(queue);
  nthreads = servers + clients;
  if (nthreads > MAXTHREADS ||
      servers <= 0 || clients <= 0) {
    usage(iam);
    return -1;
  }
  /* Start clients */
  for(thread_num=0;thread_num < clients;++thread_num) {
    threads[thread_num].tnum      = thread_num + 1;
    threads[thread_num].verbose   = verbose;
    threads[thread_num].queue     = queue;
    threads[thread_num].file      = &ofile;
    if(rv = pthread_create((pthread_t *)&threads[thread_num].tid,
    			   NULL, 
			   Client,
			   (void *)&threads[thread_num])) {
      fprintf(stderr, "Client thread %d create failed:%s\n",
	      thread_num+1, strerror(rv));
      /* Set last good thread for join */
      RunFlag = 0;
      pthread_cond_broadcast(&queue->cond);
      break;
    }
  }
  /* Start servers */
  for(i=0;i < servers;++i, ++thread_num) {
    threads[thread_num].tnum      = thread_num + 1;
    threads[thread_num].verbose   = verbose;
    threads[thread_num].queue     = queue;
    threads[thread_num].file      = ifile;
    if(rv = pthread_create((pthread_t *)&threads[thread_num].tid,
    			   NULL, 
			   Server,
			   (void *)&threads[thread_num])) {
      fprintf(stderr, "Server thread %d create failed:%s\n",
	      thread_num+1, strerror(rv));
      /* Set last good thread for join */
      RunFlag = 0;
      pthread_cond_broadcast(&queue->cond);
      break;
    }
  }
  if (RunFlag == 0) {
    fprintf(stderr, "%d threads created, error, run is false\n", 
	    thread_num);
  } else if (verbose) {
    fprintf(stderr, "%d servers, %d clients running\n", 
	    servers, clients, thread_num);
  }
  /* now wait for join */
  for(i=0;i<thread_num;i++) {
    rv = pthread_join(threads[i].tid, &trv);
    if (rv) {
      fprintf(stderr, "Thread %d join failed:%s\n",
	      threads[i].tid, strerror(rv));
    } else {
      rv = ((TARG *)trv)->rv;
      if (rv < 0) {
	fprintf(stderr, "Thread %d returned %d\n", threads[i].tnum, rv);
      } else if (verbose) {
	fprintf(stderr, "join success on %d\n", threads[i].tnum);
      }
    }
  }
  QueueDestroy(queue);
  return 0;
}
